from decimal import Decimal, ROUND_HALF_DOWN, ROUND_FLOOR

def round2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).quantize(Decimal('1.'), rounding=ROUND_HALF_DOWN) * tick_size
  return r.normalize()

def trunc2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).to_integral_value(rounding=ROUND_FLOOR) * tick_size
  return r.normalize()


from typing_extensions import Protocol, AsyncIterable, Sequence
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime
from trading_sdk.types import Side

@dataclass
class Trade:
  id: str
  price: Decimal
  qty: Decimal
  time: datetime
  side: Side
  maker: bool | None = None

class MyTrades(Protocol):
  def my_trades(
    self, symbol: str, *, limit: int | None = None,
    start: datetime | None = None, end: datetime | None = None
  ) -> AsyncIterable[Sequence[Trade]]:
    """Fetch trades (from your account) on a given symbol. Automatically paginates if needed.
    
    - `symbol`: The symbol being traded, e.g. `BTCUSDT`
    - `limit`: The maximum number of trades to fetch in a single request.
    - `start`: The start time to query. If given, only trades after this time will be returned.
    - `end`: The end time to query. If given, only trades before this time will be returned.

    ### Ordering

    - If `start` is given, trades are ordered forwards by time.
    - Otherwise, ordered backwards by time.
    """
    ...
  
  async def my_trades_seq(
    self, symbol: str, *, limit: int | None = None,
    start: datetime | None = None, end: datetime | None = None
  ) -> Sequence[Trade]:
    """Fetch trades (from your account) on a given symbol. Automatically paginates if needed.
    
    - `symbol`: The symbol being traded, e.g. `BTCUSDT`
    - `limit`: The maximum number of trades to fetch in a single request.
    - `start`: The start time to query. If given, only trades after this time will be returned.
    - `end`: The end time to query. If given, only trades before this time will be returned.

    ### Ordering

    - If `start` is given, trades are ordered forwards by time.
    - Otherwise, ordered backwards by time.
    """
    trades = []
    async for page in self.my_trades(symbol, limit=limit, start=start, end=end):
      trades.extend(page)
    return trades


from typing_extensions import Protocol
from dataclasses import dataclass
from decimal import Decimal

@dataclass
class Book:
  @dataclass
  class Entry:
    price: Decimal
    qty: Decimal

    def __str__(self) -> str:
      return f'{self:f}'

    def __format__(self, fmt: str) -> str:
      return f'[{self.price:{fmt}}] {self.qty:{fmt}}'

  bids: list[Entry]
  asks: list[Entry]

  @property
  def best_bid(self) -> Entry:
    return max(self.bids, key=lambda e: e.price)

  @property
  def best_ask(self) -> Entry:
    return min(self.asks, key=lambda e: e.price)

  def __str__(self) -> str:
    return f'{self:f}'

  def __format__(self, fmt: str) -> str:
    hr = '-'*16
    asks = sorted(self.asks, key=lambda e: e.price, reverse=True)
    bids = sorted(self.bids, key=lambda e: e.price, reverse=True)
    return (
      f'{hr}\nASKS\n{hr}\n' +
      '\n'.join(f'{e:{fmt}}' for e in asks) +
      f'\n{hr}\nBIDS\n{hr}\n' +
      '\n'.join(f'{e:{fmt}}' for e in bids) +
      f'\n{hr}'
    )

class Depth(Protocol):
  async def depth(self, symbol: str, *, limit: int | None = None) -> Book:
    """Get the order book for a given symbol.
    
    - `symbol`: The symbol being traded, e.g. `BTCUSDT`.
    - `limit`: The maximum number of bids/asks to return.
    """
    ...

  async def best_bid(self, symbol: str) -> Decimal:
    """Get the best bid (price) for a given symbol."""
    book = await self.depth(symbol, limit=1)
    return book.best_bid.price


from typing_extensions import Protocol
from trading_sdk.types import Num

class EditOrder(Protocol):
  async def edit_order(self, symbol: str, *, id: str, qty: Num) -> str:
    """Edit an existing order.
    
    - `symbol`: The symbol to edit the order for.
    - `id`: The ID of the order to edit.
    - `quantity`: The new quantity of the order.

    Returns the new ID of the order.
    """
    ...


from typing_extensions import Protocol
from trading_sdk.spot.user_data.query_order import OrderState

class CancelOrder(Protocol):
  async def cancel_order(self, symbol: str, *, id: str) -> OrderState:
    """Cancel an order.
    
    - `symbol`: The symbol to cancel the order for.
    - `id`: The ID of the order to cancel.
    """
    ...


from typing_extensions import Protocol, AsyncIterable
from trading_sdk.spot.market_data.depth import Book

class Depth(Protocol):
  def depth(self, symbol: str, *, limit: int | None = None) -> AsyncIterable[Book]:
    ...


from typing_extensions import Protocol, Sequence
from .query_order import OrderState

class OpenOrders(Protocol):
  async def open_orders(self, symbol: str) -> Sequence[OrderState]:
    """Fetch currently open orders (of your account) on a given symbol.
    
    - `symbol`: The symbol being traded, e.g. `BTCUSDT`
    """
    ...


from typing_extensions import Protocol, TypedDict, NotRequired

class Address(TypedDict):
  address: str
  memo: NotRequired[str | None]

class DepositAddress(Protocol):
  async def deposit_address(self, currency: str, *, network: str) -> Address:
    """Get the deposit address for a currency.
    
    - `currency`: The currency to get the deposit address for.
    - `network`: The network to get the deposit address for.
    """
    ...


from ._user_data import UserData
from .balances import Balances
from .my_trades import MyTrades
from .open_orders import OpenOrders
from .query_order import QueryOrder
from .query_orders import QueryOrders

__all__ = [
  'UserData',
  'Balances',
  'MyTrades',
  'OpenOrders',
  'QueryOrder',
  'QueryOrders',
]


from ._trading import Trading
from .cancel_order import CancelOrder
from .edit_order import EditOrder
from .place_order import PlaceOrder
from .place_orders import PlaceOrders

__all__ = [
  'CancelOrder',
  'PlaceOrder',
  'EditOrder',
  'Trading',
  'PlaceOrders',
]


from typing_extensions import Sequence, Protocol
import asyncio
from .place_order import Order, PlaceOrder


class PlaceOrders(PlaceOrder, Protocol):
  async def place_orders(self, symbol: str, orders: Sequence[Order]) -> Sequence[str]:
    """Place multiple orders on a given symbol.
    
    - `symbol`: The symbol to place the orders for.
    - `orders`: The orders to place.

    Returns the order IDs.
    """
    return await asyncio.gather(*[self.place_order(symbol, order) for order in orders])


from typing_extensions import Protocol, TypeVar, Mapping
from dataclasses import dataclass
from decimal import Decimal
from trading_sdk.util import trunc2tick

S = TypeVar('S', bound=str)

@dataclass
class Info:
  base: str
  """Code of the base asset."""
  quote: str
  """Code of the quote asset."""
  tick_size: Decimal
  """Tick size of the price (in quote units)."""
  step_size: Decimal
  """Step size of the quantity (in base units)."""
  min_qty_: Decimal | None = None
  """Minimum quantity of the order (in base units)."""
  max_qty: Decimal | None = None
  """Maximum quantity of the order (in base units)."""
  min_price: Decimal | None = None
  """Minimum price of the order (in quote units)."""
  max_price: Decimal | None = None
  """Maximum price of the order (in quote units)."""

  @property
  def min_qty(self) -> Decimal:
    """Minimum quantity of the order (in base units)."""
    return self.min_qty_ or self.step_size
  
  def trunc_qty(self, base_qty: Decimal) -> Decimal | None:
    """Truncate the (base asset) quantity to the nearest step size. Returns `None` if the quantity is too small."""
    qty = trunc2tick(base_qty, self.step_size)
    if qty >= self.min_qty:
      return qty
  
  def amount2qty(self, quote_amount: Decimal, *, price: Decimal) -> Decimal | None:
    """Convert a quote amount to a base quantity, truncating to the nearest step size. Returns `None` if the quantity is too small."""
    return self.trunc_qty(quote_amount / price)
  
  def qty2amount(self, base_qty: Decimal, *, price: Decimal) -> Decimal:
    """Convert a base quantity to a quote amount."""
    return base_qty * price
  
class ExchangeInfo(Protocol):
  async def exchange_info(self, symbol: str) -> Info:
    """Get the exchange info for the given symbol."""
    return (await self.exchange_infos(symbol))[symbol]

  async def exchange_infos(self, *symbols: S) -> Mapping[S, Info]:
    """Get the exchange info for the given symbols."""
    ...


from .wallet import Wallet
from .spot import Spot
from .types import Side, TimeInForce, fmt_num, Error, NetworkFailure, ApiError

__all__ = [
  'Wallet',
  'Spot',
  'Side',
  'TimeInForce',
  'fmt_num',
  'Error', 'NetworkFailure', 'ApiError',
]


from typing_extensions import Protocol, Literal
from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime
from trading_sdk.types import Side

OrderStatus = Literal['NEW', 'PARTIALLY_FILLED', 'FILLED', 'CANCELED', 'PARTIALLY_CANCELED']

@dataclass
class OrderState:
  id: str
  price: Decimal
  qty: Decimal
  filled_qty: Decimal
  side: Side
  status: OrderStatus
  time: datetime

  @property
  def quote_qty(self) -> Decimal:
    return self.qty * self.price

  @property
  def unfilled_qty(self) -> Decimal:
    return self.qty - self.filled_qty
  
class QueryOrder(Protocol):
  async def query_order(self, symbol: str, *, id: str) -> OrderState:
    """Query an order.
    
    - `symbol`: The symbol to query the order for.
    - `id`: The ID of the order to query.
    """
    ...


from typing_extensions import Protocol, TypedDict, Sequence
from decimal import Decimal

class WithdrawalMethod(TypedDict):
  network: str
  contract_address: str
  fee: Decimal

class WithdrawalMethods(Protocol):
  async def withdrawal_methods(self, currency: str) -> Sequence[WithdrawalMethod]:
    """Get the withdrawal methods for a currency.
    
    - `currency`: The currency to get the withdrawal methods for.
    """
    ...


from typing_extensions import Protocol
from .edit_order import EditOrder
from .cancel_order import CancelOrder
from .place_orders import PlaceOrders

class Trading(CancelOrder, EditOrder, PlaceOrders, Protocol):
  ...


from typing_extensions import Protocol, AsyncIterable
from trading_sdk.spot.user_data.my_trades import Trade

class MyTrades(Protocol):
  def my_trades(self, symbol: str) -> AsyncIterable[Trade]:
    ...


from typing_extensions import Protocol
from trading_sdk.types import Num

class Withdraw(Protocol):
  async def withdraw(
    self, currency: str, *, address: str, amount: Num,
    network: str | None = None, memo: str | None = None,
    contract_address: str | None = None,
  ):
    """Withdraw funds.
    
    - `currency`: The currency to withdraw.
    - `address`: The address to withdraw to.
    - `amount`: The amount to withdraw.
    - `network`: The network to withdraw from.
    - `memo`: The memo to withdraw to (for some networks).
    - `contract_address`: The contract address to withdraw to. You can use it to make sure it's the token you expect.
    """
    ...


from typing_extensions import Protocol, Mapping, TypeVar
from dataclasses import dataclass
from decimal import Decimal

S = TypeVar('S', bound=str, default=str)

@dataclass
class Balance:
  free: Decimal
  locked: Decimal

  @property
  def total(self) -> Decimal:
    return self.free + self.locked

class Balances(Protocol):

  async def balance(self, currency: S, /) -> Balance:
    """Get the balance of the given currency."""
    return (await self.balances(currency))[currency]

  async def balances(self, *currencies: S) -> Mapping[S, Balance]:
    """Get the balances of the given currencies."""
    ...


from typing_extensions import Protocol, TypedDict, Literal
from trading_sdk.types import Side, Num

class BaseOrder(TypedDict):
  side: Side
  qty: Num
  """Quantity of the order in the base asset."""

class LimitOrder(BaseOrder):
  price: Num
  type: Literal['LIMIT']

class MarketOrder(BaseOrder):
  type: Literal['MARKET']

Order = LimitOrder | MarketOrder

class PlaceOrder(Protocol):
  async def place_order(self, symbol: str, order: Order) -> str:
    """Place an order.
    
    - `symbol`: The symbol to place the order for.
    - `order`: The order to place.

    Returns the order ID.
    """
    ...


from typing_extensions import Literal, Any
from dataclasses import dataclass
from decimal import Decimal

Side = Literal['BUY', 'SELL']
TimeInForce = Literal['GTC', 'IOC', 'FOK']
Num = Decimal | str | float | int

def fmt_num(num: Num) -> str:
  if isinstance(num, str):
    return num
  else:
    return f'{num:f}'
  
class Error(Exception):
  ...
  
@dataclass
class NetworkFailure(Error):
  detail: Any

@dataclass
class ApiError(Error):
  detail: Any


from ._spot import Spot
from .market_data import MarketData, Depth, Time, ExchangeInfo
from .user_data import UserData, MyTrades, OpenOrders, QueryOrder, QueryOrders, Balances
from .trading import Trading, CancelOrder, EditOrder, PlaceOrders, PlaceOrder

__all__ = [
  'Spot',
  'MarketData',
  'Depth',
  'Time',
  'ExchangeInfo',
  'UserData',
  'Trading',
  'CancelOrder',
  'EditOrder',
  'PlaceOrders',
  'PlaceOrder',
  'MyTrades',
  'OpenOrders',
  'QueryOrder',
  'QueryOrders',
  'Balances',
]


import asyncio
from typing_extensions import Sequence, Protocol
from .query_order import QueryOrder, OrderState

class QueryOrders(QueryOrder, Protocol):
  async def query_orders(self, symbol: str, *, ids: Sequence[str]) -> Sequence[OrderState]:
    """Query multiple orders by symbol and ID.
    
    - `symbol`: The symbol to query orders for.
    - `ids`: The IDs of the orders to query.

    Returns the states of the orders.
    """
    return await asyncio.gather(*[self.query_order(symbol, id=id) for id in ids])


from typing_extensions import Protocol
from .balances import Balances
from .my_trades import MyTrades
from .open_orders import OpenOrders
from .query_orders import QueryOrders

class UserData(Balances, MyTrades, OpenOrders, QueryOrders, Protocol):
  ...


from ._wallet import Wallet
from .deposit_address import DepositAddress
from .withdraw import Withdraw
from .withdrawal_methods import WithdrawalMethods

__all__ = ['Wallet', 'DepositAddress', 'Withdraw', 'WithdrawalMethods']


from typing_extensions import Protocol
from .deposit_address import DepositAddress
from .withdraw import Withdraw
from .withdrawal_methods import WithdrawalMethods

class Wallet(DepositAddress, Withdraw, WithdrawalMethods, Protocol):
  ...